from abc import ABC, abstractmethod

from metripy.Application.Config.ProjectConfig import ProjectConfig
from metripy.Component.Output.ProgressBar import ProgressBar
from metripy.LangAnalyzer.Generic.CodeSmell.CodeSmellDetectorFactory import (
    CodeSmellDetectorFactory,
)
from metripy.LangAnalyzer.Generic.DuplicateSearch.DuplicateDetector import (
    DuplicateDetector,
)
from metripy.LangAnalyzer.Generic.DuplicateSearch.TokenizerFactory import (
    TokenizerFactory,
)
from metripy.LangAnalyzer.Generic.Metrics.GenericHalSteadAnalyzer import (
    GenericHalSteadAnalyzer,
)
from metripy.LangAnalyzer.Generic.Metrics.GenericLocAnalyzer import GenericLocAnalyzer
from metripy.LangAnalyzer.Generic.Metrics.HalSteadAnalyzerFactory import (
    HalSteadAnalyzerFactory,
)
from metripy.LangAnalyzer.Generic.Metrics.LocAnalyzerFactory import LocAnalyzerFactory
from metripy.Metric.Code.FileMetrics import FileMetrics
from metripy.Tree.FunctionNode import FunctionNode
from metripy.Tree.ModuleNode import ModuleNode


class AbstractLangAnalyzer(ABC):
    def __init__(self, project_config: ProjectConfig):
        self.config = project_config
        self.files: list[str] = []
        self.modules: dict[str, ModuleNode] = {}
        self.loc_analyzer: GenericLocAnalyzer = LocAnalyzerFactory.get_loc_analyzer(
            self.get_lang_name()
        )
        self.halstead_analyzer: GenericHalSteadAnalyzer = (
            HalSteadAnalyzerFactory.get_halstead_analyzer(self.get_lang_name())
        )
        self.code_smell_detector = CodeSmellDetectorFactory.get_code_smell_detector(
            self.get_lang_name(), self.config.code_smells
        )
        self.duplicate_detector = DuplicateDetector(
            tokenizer=TokenizerFactory.get_tokenizer(self.get_lang_name())
        )

    # need to be implemented by sub analyzers
    @abstractmethod
    def get_lang_name(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_supported_extensions(self) -> tuple[str]:
        raise NotImplementedError

    # can be overridden by sub analyzers
    def before_run(self) -> None:
        # build cache
        pass

    def after_run(self) -> None:
        # clear cache
        pass

    def set_files(self, files: list[str]) -> None:
        self.files = list(
            filter(lambda file: file.endswith(self.get_supported_extensions()), files)
        )

    def is_needed(self) -> bool:
        return len(self.files) > 0

    def run(self, progress_bar: ProgressBar) -> None:
        for file in self.files:
            with open(file, "r") as f:
                code = f.read()
                self.analyze(code, file)
            progress_bar.advance()

    def create_module_node(self, filename: str, code: str) -> ModuleNode:
        loc_data = self.loc_analyzer.analyze(code)
        full_name = self.full_name(filename)
        return ModuleNode(
            full_name,
            loc_data.get("loc", 0),
            loc_data.get("lloc", 0),
            loc_data.get("sloc", 0),
            loc_data.get("comments", 0),
            loc_data.get("multiline_comments", 0),
            loc_data.get("blank_lines", 0),
            loc_data.get("single_comments", 0),
        )

    def add_function_halstead_metrics(
        self, function_node: FunctionNode, function_code: str
    ) -> None:
        function_metrics = self.halstead_analyzer.calculate_halstead_metrics(
            function_code
        )
        function_node.h1 = function_metrics["n1"]
        function_node.h2 = function_metrics["n2"]
        function_node.N1 = function_metrics["N1"]
        function_node.N2 = function_metrics["N2"]
        function_node.vocabulary = function_metrics["vocabulary"]
        function_node.length = function_metrics["length"]
        function_node.calculated_length = function_metrics["calculated_length"]
        function_node.volume = function_metrics["volume"]
        function_node.difficulty = function_metrics["difficulty"]
        function_node.effort = function_metrics["effort"]
        function_node.bugs = function_metrics["bugs"]
        function_node.time = function_metrics["time"]
        function_node.calc_mi()

    @staticmethod
    def full_name(
        filename: str, item_name: str | None = None, class_name: str | None = None
    ) -> str:
        if class_name is None:
            if item_name is None:
                return filename
            return f"{filename}:{item_name}"
        return f"{filename}:{class_name}:{item_name}"

    def get_metrics(self) -> list[FileMetrics]:
        metrics: dict[str, FileMetrics] = {}

        efferent_coupling = {}
        afferent_coupling = {}

        for module in self.modules.values():
            full_name = module.full_name

            if len(module.functions) > 0:
                totalCc = sum(function.complexity for function in module.functions)
                avgCcPerFunction = totalCc / len(module.functions)
                avgLocPerFunction = sum(
                    function.get_loc() for function in module.functions
                ) / len(module.functions)
                total_cog_complexity = sum(
                    function.cognitive_complexity for function in module.functions
                )
                avg_cog_complexity_per_function = total_cog_complexity / len(
                    module.functions
                )
            else:
                totalCc = 0
                avgCcPerFunction = 0
                avgLocPerFunction = 0
                total_cog_complexity = 0
                avg_cog_complexity_per_function = 0
            maintainabilityIndex = module.maintainability_index

            file_metric = FileMetrics(
                full_name=full_name,
                loc=module.loc,
                totalCc=totalCc,
                avgCcPerFunction=avgCcPerFunction,
                maintainabilityIndex=maintainabilityIndex,
                avgLocPerFunction=avgLocPerFunction,
                class_nodes=module.classes,
                function_nodes=module.functions,
                import_name=module.import_name,
                imports=module.imports,
                code_smells=module.code_smells,
                total_cog_complexity=total_cog_complexity,
                avg_cog_complexity_per_function=avg_cog_complexity_per_function,
            )
            metrics[full_name] = file_metric

            if not module.import_name or not module.imports:
                continue
            efferent_coupling[module.import_name] = len(module.imports)
            for import_name in module.imports:
                if import_name not in afferent_coupling:
                    afferent_coupling[import_name] = []
                afferent_coupling[import_name].append(module.import_name)

        for file_metric in metrics.values():
            imported_by = afferent_coupling.get(file_metric.import_name, [])
            ca = len(imported_by)
            ce = efferent_coupling.get(file_metric.import_name, 0)
            file_metric.imported_by = imported_by
            file_metric.afferent_coupling = ca
            file_metric.efferent_coupling = ce
            file_metric.instability = (ce / (ca + ce)) if (ca + ce) > 0 else 0

        return list(metrics.values())



import math
import os
import re
import tempfile
from pathlib import Path

import lizard

from metripy.Application.Config.ProjectConfig import ProjectConfig
from metripy.LangAnalyzer.AbstractLangAnalyzer import AbstractLangAnalyzer
from metripy.LangAnalyzer.Php.Ast.PhpAstParser import PhpAstParser
from metripy.LangAnalyzer.Php.Metrics.PhpCognitiveComplexityCalculator import (
    PhpCognitiveComplexityCalculator,
)
from metripy.LangAnalyzer.Php.PhpBasicAstParser import PhpBasicAstParser
from metripy.LangAnalyzer.Php.PhpImportsAnalyzer import PhpImportsAnalyzer
from metripy.Tree.ClassNode import ClassNode
from metripy.Tree.FunctionNode import FunctionNode
from metripy.Tree.ModuleNode import ModuleNode


class PhpAnalyzer(AbstractLangAnalyzer):
    def __init__(self, project_config: ProjectConfig):
        super().__init__(project_config)

    def get_lang_name(self) -> str:
        return "PHP"

    def get_supported_extensions(self) -> tuple[str]:
        return (".php",)

    def _create_lizard_analyzable_file(self, filename: str) -> str:
        """
        Because of a bug in lizard it cannot correctly analyze traits.
        See https://github.com/terryyin/lizard/issues/441
        Because of a bug in lizard we need to replace use function statements.
        See https://github.com/terryyin/lizard/issues/442
        """
        tmp_file = tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".php")
        with open(filename, "r") as f:
            code = f.readlines()
            code = [re.sub(r"^(\s*)trait\s+", r"\1class ", line) for line in code]
            # use remove the function keyword, so it gets counted the same way and dont increate blank lines / lower loc
            code = [
                re.sub(r"^(\s*)use\s+function\s+", r"\1use ", line) for line in code
            ]
            tmp_file.writelines(code)
            tmp_file.close()
        return tmp_file.name

    def analyze(self, code: str, filename: str) -> None:
        file_stem = Path(filename).stem
        structure = PhpBasicAstParser.parse_php_structure(code)

        is_tmp_file = False
        filename_to_analyze = filename
        if "trait" in filename.lower() or "use function" in code:
            filename_to_analyze = self._create_lizard_analyzable_file(filename)
            is_tmp_file = True

        lizard_result = lizard.analyze_file(filename_to_analyze)
        complexity_data = {
            func.name: {
                "complexity": func.cyclomatic_complexity,
                "start_line": func.start_line,
                "end_line": func.end_line,
            }
            for func in lizard_result.function_list
        }

        if is_tmp_file:
            os.unlink(filename_to_analyze)

        classes: dict[str, ClassNode] = {}
        functions: dict[str, FunctionNode] = {}
        for obj in structure:
            if obj["type"] == "class":
                full_name = self.full_name(filename, obj["name"])
                classes[full_name] = ClassNode(
                    full_name,
                    obj["name"],
                    obj["line"],
                    0,
                    0,  # gets filled in later, based on methods
                )
            elif obj["type"] == "method" or obj["type"] == "function":
                full_name = self.full_name(filename, obj["name"])
                try:
                    function_node = FunctionNode(
                        full_name,
                        obj["name"],
                        obj["line"],
                        0,
                        complexity_data[f"{file_stem}::{obj['name']}"]["complexity"],
                    )
                    function_node.line_end = complexity_data[
                        f"{file_stem}::{obj['name']}"
                    ]["end_line"]
                except KeyError:
                    # no complexity data, function must be empty
                    function_node = FunctionNode(
                        full_name,
                        obj["name"],
                        obj["line"],
                        0,
                        0,
                    )

                if obj["type"] == "method":
                    class_name = obj["class"]
                    full_class_name = self.full_name(filename, class_name)
                    class_node = classes.get(full_class_name)
                    if class_node is not None:
                        class_node.functions.append(function_node)
                    else:
                        raise ValueError(
                            f"Class node not found for function {full_class_name}"
                        )
                functions[full_name] = function_node

        # complexity of classes
        for class_node in classes.values():
            class_node.real_complexity = sum(
                [func.complexity for func in class_node.functions]
            )

        module_node = self.create_module_node(filename, code)
        module_node.classes.extend(classes.values())
        module_node.functions.extend(functions.values())

        code_lines = code.split("\n")
        for func_name, function_node in functions.items():
            lines = code_lines[function_node.lineno : function_node.line_end]
            self.add_function_halstead_metrics(function_node, "\n".join(lines))

        maintainability_index = self._calculate_maintainability_index(
            functions.values(), module_node
        )
        module_node.maintainability_index = maintainability_index
        self.modules[module_node.full_name] = module_node

        # ignore for now
        # self.duplicate_detector.add_code(filename, code)

        imports_analyzer = PhpImportsAnalyzer(filename, code)
        module_node.import_name = imports_analyzer.extract_import_name()
        module_node.imports = imports_analyzer.extract_imports()

        module_node.code_smells = self.code_smell_detector.detect_all(filename, code)

        parser = PhpAstParser()
        parser.parse(code)
        cognitive_complexities = PhpCognitiveComplexityCalculator(
            parser
        ).calculate_for_all_functions()
        for func_name, complexity in cognitive_complexities.items():
            full_name = self.full_name(filename, func_name)
            function_node = functions.get(full_name)
            if function_node is not None:
                function_node.cognitive_complexity = complexity
            else:
                raise ValueError(f"Function node not found for function {full_name}")

    def _calculate_maintainability_index(
        self, functions: list[FunctionNode], module_node: ModuleNode
    ) -> float:
        """Calculate maintainability index for PHP"""
        if not functions:
            return 100.0

        total_volume = sum(func.volume for func in functions)
        total_complexity = sum(func.complexity for func in functions)
        total_length = sum(func.length for func in functions)

        if total_volume == 0 or total_length == 0:
            return 100.0

        # PHP maintainability index calculation
        mi_base = max(
            (
                171
                - 5.2 * math.log(total_volume)
                - 0.23 * total_complexity
                - 16.2 * math.log(total_length)
            )
            * 100
            / 171,
            0,
        )

        # Comment weight
        comment_weight = 0
        if module_node.loc > 0:
            comment_ratio = module_node.single_comments / module_node.loc
            comment_weight = 50 * math.sin(math.sqrt(2.4 * comment_ratio))

        return mi_base + comment_weight
